# -*- coding: utf-8 -*-
##############################################################################
#
#    wms_inventory module for OpenERP, Module to extended inventory process
#    Copyright (C) 2010 SYLEAM (<http://www.syleam.fr/>)
#              Christophe CHAUVET <christophe.chauvet@syleam.fr>
#
#    This file is a part of wms_inventory
#
#    wms_inventory is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    wms_inventory is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from osv import osv
from osv import fields
from osv.orm import except_orm
from tools.translate import _
import netsvc
import thread
import pooler
import time
import tempfile
import base64
import time


class Inventory(osv.osv):
    _inherit = 'stock.inventory'

    _columns = {
        'type': fields.selection([('global', 'Global'), ('partial', 'Partial')], 'Type',
            help='If global, reset all quantity product  not found in this inventory'),
        'warehouse_id': fields.many2one('stock.warehouse', 'Warehouse', required=True,
            help='Select the warehouse where the inventory is made'),
        'state': fields.selection([('draft', 'Draft'), ('progress', 'Progress'), ('done', 'Done')], 'Status', readonly=True),
    }

    _defaults = {
        'type': lambda *a: 'partial',
        'warehouse_id': lambda obj, cr, uid, context: obj.pool.get('res.users')._default_warehouse(cr, uid, context),
        'state': lambda *a: 'draft',
    }

    def copy(self, cr, uid, id, default=None, context=None):
        """
        When copy the inventory, don't copy the stock_move
        """
        if context is None:
            context = {}

        default['move_ids'] = []

        return super(Inventory, self).copy(cr, uid, id, default, context=context)

    def action_done(self, cr, uid, ids, context=None):
        """
        Validate the inventory, it take a lot of time and must be launch in multi thread
        """
        if context is None:
            context = {}

        def threading_done(cr, uid, ids, context=None):
            """
            Treatment launch as a new thread
            Must be include production lot and tracking
            """
            def log_message(fic, section, message):
                """
                Format the content of the logfile
                """
                fic.write('[%s] %s: %s\n' % (time.strftime('%Y-%m-%d %H:%M:%S'), section.encode('utf-8'), message.encode('utf-8')))

            logger = netsvc.Logger()
            logger.notifyChannel('wms_inventory', netsvc.LOG_INFO, '[%s] Start compute the inventory *********' % cr.dbname)
            cr = pooler.get_db(cr.dbname).cursor()
            for inv in self.browse(cr, uid, ids):
                logger.notifyChannel('wms_inventory', netsvc.LOG_INFO, '[%s] Inventory %s' % (cr.dbname, inv.name))
                # For each inventory, we save all step in the log file
                outfp = tempfile.NamedTemporaryFile(prefix='openerp', suffix='inventory')
                log_message(outfp, _('init'), _('Inventory number %s') % inv.name)
                if inv.type == 'partial':
                    log_message(outfp, _('init'), _(' -> Partial Inventory'))
                else:
                    log_message(outfp, _('init'), _(' -> Global Inventory'))
                inv_message = ''
                # If default inventory location is missing in the warehouse, we add information in the logfile and exit properly
                if inv.warehouse_id.lot_inventory_id:
                    log_message(outfp, _('init'), _('Default inventory location %s') % inv.warehouse_id.lot_inventory_id.name)
                    move_ids = []
                    move_line = []
                    inv_ok = True
                    tmp_prod = []
                    log_message(outfp, 'init', _('Start to analyse each line'))
                    for line in inv.inventory_line_id:
                        log_message(outfp, _('analyze'), '**********************')
                        pid = line.product_id.id
                        log_message(outfp, _('analyze'), _('product %s') % line.product_id.name)
                        log_message(outfp, _('analyze'), _('        [%s]') % line.product_id.default_code)
                        price = line.product_id.standard_price or 0.0
                        log_message(outfp, _('analyze'), _('Cost price %.2f') % line.product_id.standard_price)
                        prodlot = line.prodlot_id and line.prodlot_id.id or False
                        if prodlot:
                            log_message(outfp, _('analyze'), _('production lot %s') % line.prodlot_id.name)
                        tracking = line.tracking_id and line.tracking_id.id or False
                        if tracking:
                            log_message(outfp, _('analyze'), _('tracking %s') % line.tracking_id.name)

                        amount = self.pool.get('stock.location')._stock_get(cr, uid, line.location_id.id, [pid],
                                {'uom': line.product_uom.id, 'prodlot': prodlot, 'tracking': tracking})
                        log_message(outfp, _('analyze'), _('quantity in stock %.2f at location %s') % (amount, line.location_id.name))
                        log_message(outfp, _('analyze'), _('quantity count %.2f') % line.product_qty)

                        # for global inventory, we keep if we found this product
                        tmp_prod.append('%d_%d_%d_%d' % (line.product_id.id, line.location_id.id, prodlot or 0, tracking or 0))
                        change = line.product_qty - amount
                        if not change:
                            log_message(outfp, _('action'), _('nothing to regulate'))
                        else:
                            log_message(outfp, _('action'), _('quantity to regulate %.2f') % change)
                            location_id = line.product_id.product_tmpl_id.property_stock_inventory.id or inv.warehouse_id.lot_inventory_id.id
                            value = {
                                'name': 'INV:' + str(line.inventory_id.id) + ':' + line.inventory_id.name,
                                'product_id': line.product_id.id,
                                'product_uom': line.product_uom.id,
                                'date': inv.date,
                                'date_planned': inv.date,
                                'state': 'assigned',
                                'prodlot_id': prodlot,
                                'tracking_id': tracking,
                            }
                            if change > 0:
                                value.update({
                                    'product_qty': change,
                                    'location_id': location_id,
                                    'location_dest_id': line.location_id.id,
                                })
                            else:
                                value.update({
                                    'product_qty': -change,
                                    'location_id': line.location_id.id,
                                    'location_dest_id': location_id,
                                })
                            try:
                                move_ids.append(self.pool.get('stock.move').create(cr, uid, value))
                            except except_orm, e:
                                logger.notifyChannel('wms_inventory', netsvc.LOG_INFO, '[%s] %s %s' % (cr.dbname, e.name, e.value))
                                log_message(outfp, _('error'), '[%s] %s %s' % (cr.dbname, e.name, e.value))
                                inv_message += '    %s: %s\n' % (e.name, e.value)
                                inv_ok = False

                    ##
                    ## Global inventory must reset the missing product, on this warehouse
                    log_message(outfp, _('global'), '===========================================')
                    if inv.type == 'global':
                        avail_obj = self.pool.get('wms.report.stock.available')
                        args = [
                            ('warehouse_id', '=', inv.warehouse_id.id),
                        ]
                        avail_ids = avail_obj.search(cr, uid, args, context=context)
                        for a in avail_obj.browse(cr, uid, avail_ids, context=context):
                            # Check if this product was already found at this location
                            tmp_key = '%d_%d_%d_%d' % (a.product_id.id, a.location_id.id,
                                        a.prodlot_id and a.prodlot_id.id or 0, a.tracking_id and a.tracking_id.id or 0)

                            if tmp_key in tmp_prod:
                                continue

                            log_message(outfp, _('analyze'), '**********************')
                            log_message(outfp, _('analyze'), _('product %s') % a.product_id.name)
                            log_message(outfp, _('analyze'), _('        [%s]') % a.product_id.default_code)
                            if a.prodlot_id:
                                log_message(outfp, _('analyze'), _('production lot %s') % a.prodlot_id.name)
                            if a.tracking_id:
                                log_message(outfp, _('analyze'), _('tracking %s') % a.tracking_id.name)
                            log_message(outfp, _('action'), _('quantity to reset %.2f') % a.qty)

                            location_id = line.product_id.product_tmpl_id.property_stock_inventory.id or inv.warehouse_id.lot_inventory_id.id
                            value = {
                                'name': 'INV: reset qty: ' + line.inventory_id.name,
                                'product_id': a.product_id.id,
                                'product_uom': a.uom_id.id,
                                'date': inv.date,
                                'date_planned': inv.date,
                                'state': 'assigned',
                                'prodlot_id': a.prodlot_id and a.prodlot_id.id or False,
                                'tracking_id': a.tracking_id and a.tracking_id.id or False,
                                'product_qty': a.qty,
                                'location_id': a.location_id.id,
                                'location_dest_id': location_id,
                            }
                            try:
                                move_ids.append(self.pool.get('stock.move').create(cr, uid, value))
                            except except_orm, e:
                                logger.notifyChannel('wms_inventory', netsvc.LOG_INFO, '[%s] %s %s' % (cr.dbname, e.name, e.value))
                                log_message(outfp, _('error'), '[%s] %s %s' % (cr.dbname, e.name, e.value))
                                inv_message += '    %s: %s\n' % (e.name, e.value)
                                inv_ok = False

                    logger.notifyChannel('wms_inventory', netsvc.LOG_INFO, '[%s] Compute each line' % cr.dbname)
                    log_message(outfp, _('action'), _('Validate each inventory line'))
                    if len(move_ids) and inv_ok:
                        try:
                            self.pool.get('stock.move').action_done(cr, uid, move_ids,
                                    context=context)
                        except except_orm, e:
                            logger.notifyChannel('wms_inventory', netsvc.LOG_INFO, '[%s] %s %s' % (cr.dbname, e.name, e.value))
                            inv_message += '    %s: %s\n' % (e.name, e.value)
                            inv_ok = False

                    log_message(outfp, _('action'), _('Finish the current inventory'))
                    if inv_ok:
                        self.write(cr, uid, [inv.id], {'state': 'done', 'date_done': time.strftime('%Y-%m-%d %H:%M:%S'),
                            'move_ids': [(6, 0, move_ids)]})
                else:
                    log_message(outfp, 'error', 'defaut inventory location is missing on the warehouse %s' % inv.warehouse_id.name)
                    inv_ok = False

                ###
                ## Send a request to the user whom launch the validation
                if inv_ok:
                    body = _('Inventory %s has been finished') % inv.name
                else:
                    cr.rollback()
                    body = _('Inventory %s failed\nsee below\n') % inv.name
                    body += inv_message

                self.pool.get('res.request').create(cr, uid, {
                    'name': _('Compute inventory %s done') % inv.name,
                    'act_from': uid,
                    'act_to': uid,
                    'body': body,
                }, context=context)

                # Store the logfile as attachment
                outfp.flush()
                outfp.seek(0)
                vals = {
                    'datas': base64.encodestring(outfp.read()),
                    'datas_fname': 'inv_%s.log' % inv.name.replace(' ', '_'),
                    'name': 'INV %s' % inv.name + ' [' + time.strftime('%Y%m%d%H%M%S') + ']',
                    'res_model': 'stock.inventory',
                    'res_id': inv.id,
                }
                self.pool.get('ir.attachment').create(cr, uid, vals, context=context)
                outfp.close()

                cr.commit()
            cr.close()
            logger.notifyChannel('wms_inventory', netsvc.LOG_INFO, '[%s] End compute the inventory ***********' % cr.dbname)
            return True

        # change state fo progress
        self.write(cr, uid, ids, {'state': 'progress'}, context=context)
        thread.start_new_thread(threading_done, (cr, uid, ids, context))
        return True

Inventory()


class InventoryLine(osv.osv):
    _inherit = 'stock.inventory.line'

    _columns = {
        'prodlot_id': fields.many2one('stock.production.lot', 'Production Lot'),
        'tracking_id': fields.many2one('stock.tracking', 'Tracking'),
    }

    def on_change_product_id(self, cr, uid, ids, location_id, product, uom=False):
        if not product:
            return {}

        if not uom:
            prod = self.pool.get('product.product').browse(cr, uid, [product], {'uom': uom})[0]
            uom = prod.uom_id.id

        return {'value': {'product_uom': uom}}

InventoryLine()

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:
